@@ -25,8 +25,12 @@ from codecs import getincrementaldecoder
 
 from .code import Code
 from .inputstream import InputStream
-from .inputstream_handler import InputStreamHandler
-from .line import Line, Exit, ClearScreen, ReturnInput, Abort, ListCompletions
+from .key_binding import InputProcessor
+from .enums import InputMode
+from .key_binding import Registry
+from .key_bindings.emacs import emacs_bindings
+from .key_bindings.vi import vi_bindings
+from .line import Line, Callbacks
 from .prompt import Prompt
 from .renderer import Renderer
 from .utils import raw_mode, call_on_sigwinch
@@ -34,6 +38,9 @@ from .history import History
 
 from pygments.styles.default import DefaultStyle
 
+import weakref
+
+
 __all__ = (
         'AbortAction',
         'CommandLine',
@@ -49,6 +56,14 @@ class AbortAction:
     RETURN_NONE = 'return-none'
 
 
+class Exit(Exception):
+    pass
+
+
+class Abort(Exception):
+    pass
+
+
 class CommandLine(object):
     """
     Wrapper around all the other classes, tying everything together.
@@ -73,7 +88,8 @@ class CommandLine(object):
     inputstream_factory = InputStream
 
     #: `InputStreamHandler` class for the keybindings.
-    inputstream_handler_factory = InputStreamHandler
+    key_bindings_factories = [ emacs_bindings ]
+    default_input_mode = InputMode.VI_INSERT
 
     #: `Renderer` class.
     renderer_factory = Renderer
@@ -84,10 +100,6 @@ class CommandLine(object):
     #: `pygments.style.Style` class for the syntax highlighting.
     style = DefaultStyle
 
-    #: Boolean to indicate whether we will have other threads communicating
-    #: with the input event loop.
-    enable_concurency = False
-
     #: When to call the `on_input_timeout` callback.
     input_timeout = .5
 
@@ -97,14 +109,25 @@ class CommandLine(object):
         self.stdin = stdin or sys.stdin
         self.stdout = stdout or sys.stdout
 
-        self._line = self.line_factory(
+        #: The `Line` instance.
+        self.line = self.line_factory(
                                 code_factory=self.code_factory,
-                                history_factory=self.history_factory)
+                                history_factory=self.history_factory,
+                                callbacks=LineCallbacks(self))
 
-        self._inputstream_handler = self.inputstream_handler_factory(self._line)
+        key_registry = Registry()
+        for kb in self.key_bindings_factories:
+            kb(key_registry, self.line)
 
-        self._renderer = self.renderer_factory(
-                                prompt_factory=self.prompt_factory,
+        #: The `InputProcessor` instance.
+        self.input_processor = InputProcessor(key_registry, default_input_mode=self.default_input_mode)
+
+        #: The `prompt` instance.
+        self.prompt = self.prompt_factory(weakref.ref(self))
+
+        #: The `Renderer` instance.
+        self.renderer = self.renderer_factory(
+                                prompt=self.prompt,
                                 stdout=self.stdout,
                                 style=self.style)
 
@@ -119,12 +142,17 @@ class CommandLine(object):
         # it could be that we are in the middle of a utf-8 byte sequence.
         self._stdin_decoder = self.stdin_decoder_cls()
 
+        self._reset()
+
+    def _reset(self):
+        self._exit_flag = False
+        self._abort_flag = False
+        self._return_code = None
+
     def request_redraw(self):
         """
         Thread safe way of sending a repaint trigger to the input event loop.
         """
-        assert self.enable_concurency
-
         if self._redraw_pipe:
             os.write(self._redraw_pipe[1], b'x')
 
@@ -133,7 +161,7 @@ class CommandLine(object):
         Render the command line again. (Not thread safe!)
         (From other threads, or if unsure, use `request_redraw`.)
         """
-        self._renderer.render(self._line.get_render_context())
+        self.renderer.render()
 
     def run_in_executor(self, callback):
         """
@@ -141,15 +169,13 @@ class CommandLine(object):
         (This is recommended for code that could block the `read_input` event
         loop.)
         """
-        assert self.enable_concurency
-
         class _Thread(threading.Thread):
             def run(t):
                 callback()
 
         _Thread().start()
 
-    def on_input_timeout(self, code_obj):
+    def on_input_timeout(self):
         """
         Called when there is no input for x seconds.
         """
@@ -186,7 +212,7 @@ class CommandLine(object):
                 timeout = None
             else:
                 #
-                self.on_input_timeout(self._line.create_code_obj())
+                self.on_input_timeout()
                 timeout = None
 
     def _read_from_stdin(self):
@@ -229,18 +255,18 @@ class CommandLine(object):
 
         try:
             # Create a pipe for inter thread communication.
-            if self.enable_concurency:
-                self._redraw_pipe = os.pipe()
+            self._redraw_pipe = os.pipe()
+            fcntl.fcntl(self._redraw_pipe[0], fcntl.F_SETFL, os.O_NONBLOCK)
 
-                # Make the read-end of this pipe non blocking.
-                fcntl.fcntl(self._redraw_pipe[0], fcntl.F_SETFL, os.O_NONBLOCK)
-
-            stream = self.inputstream_factory(self._inputstream_handler, stdout=self.stdout)
+            stream = self.inputstream_factory(self.input_processor, stdout=self.stdout)
 
             def reset_line():
-                # Reset line
-                self._line.reset(initial_value=initial_value)
-                self._renderer.reset()
+                """ Reset everything. """
+                stream.reset()
+                self.line.reset(initial_value=initial_value)
+                self.renderer.reset()
+                self.input_processor.reset()
+                self._reset()
             reset_line()
 
             # Trigger read_start.
@@ -249,54 +275,48 @@ class CommandLine(object):
             with raw_mode(self.stdin):
                 self._redraw()
 
-                with call_on_sigwinch(self._redraw):
+                # When the window size changes, we do a redraw request call.
+                # (We do it asynchronously, because doing the actual drawing
+                # inside the signal handler causes easily reentrant calls,
+                # giving runtime errors.)
+                with call_on_sigwinch(self.request_redraw):
                     while True:
-                        if self.enable_concurency:
-                            c = self._get_char_loop()
-                        else:
-                            c = self._read_from_stdin()
+                        c = self._get_char_loop()
 
                         # If we got a character, feed it to the input stream. If we
                         # got none, it means we got a repaint request.
                         if c:
-                            try:
-                                # Feed one character at a time. Feeding can cause the
-                                # `Line` object to raise Exit/Abort/ReturnInput
-                                stream.feed(c)
+                            # Feed input text.
+                            stream.feed(c)
 
-                            except ClearScreen:
-                                self._renderer.clear()
+                        # If the exit flag has been set.
+                        if self._exit_flag:
+                            if on_exit != AbortAction.IGNORE:
+                                self.renderer.render(abort=True)
 
-                            except ListCompletions as e:
-                                self._renderer.render_completions(e.completions)
+                            if on_exit == AbortAction.RAISE_EXCEPTION:
+                                raise Exit()
+                            elif on_exit == AbortAction.RETURN_NONE:
+                                return None
+                            elif on_exit == AbortAction.RETRY:
+                                reset_line()
 
-                            except Exit as e:
-                                # Handle exit.
-                                if on_exit != AbortAction.IGNORE:
-                                    self._renderer.render(e.render_context)
+                        # If the abort flag has been set.
+                        if self._abort_flag:
+                            if on_abort != AbortAction.IGNORE:
+                                self.renderer.render(abort=True)
 
-                                if on_exit == AbortAction.RAISE_EXCEPTION:
-                                    raise
-                                elif on_exit == AbortAction.RETURN_NONE:
-                                    return None
-                                elif on_exit == AbortAction.RETRY:
-                                    reset_line()
+                            if on_abort == AbortAction.RAISE_EXCEPTION:
+                                raise Abort()
+                            elif on_abort == AbortAction.RETURN_NONE:
+                                return None
+                            elif on_abort == AbortAction.RETRY:
+                                reset_line()
 
-                            except Abort as abort:
-                                # Handle abort.
-                                if on_abort != AbortAction.IGNORE:
-                                    self._renderer.render(abort.render_context)
-
-                                if on_abort == AbortAction.RAISE_EXCEPTION:
-                                    raise
-                                elif on_abort == AbortAction.RETURN_NONE:
-                                    return None
-                                elif on_abort == AbortAction.RETRY:
-                                    reset_line()
-
-                            except ReturnInput as input:
-                                self._renderer.render(input.render_context)
-                                return input.document
+                        # If a return value has been set.
+                        if self._return_code:
+                            self.renderer.render(accept=True)
+                            return self._return_code
 
                         # Now render the current prompt to the output.
                         self._redraw()
@@ -315,6 +335,35 @@ class CommandLine(object):
 
             self.is_reading_input = False
 
+    def set_exit(self):
+        self._exit_flag = True
+
+    def set_abort(self):
+        self._abort_flag = True
+
+    def set_return_value(self, code):
+        self._return_code = code
+
+
+class LineCallbacks(Callbacks):
+    """
+    Callback handlers for the `Line`
+    """
+    def __init__(self, command_line):
+        self.command_line = command_line
+
+    def clear_screen(self):
+        self.command_line.renderer.clear()
+
+    def exit(self):
+        self.command_line.set_exit()
+
+    def abort(self):
+        self.command_line.set_abort()
+
+    def return_input(self, code):
+        self.command_line.set_return_value(code)
+
 
 def _select(*args, **kwargs):
     """
