@@ -15,13 +15,16 @@ from pygments.token import Keyword, Operator, Number, Name, Error, Comment, Toke
 
 from prompt_toolkit import CommandLine
 from prompt_toolkit.code import Completion, Code, ValidationError
-from prompt_toolkit.enums import LineMode
+from prompt_toolkit.enums import InputMode
 from prompt_toolkit.history import FileHistory, History
-from prompt_toolkit.inputstream_handler import ViInputStreamHandler, EmacsInputStreamHandler, ViMode
+from prompt_toolkit.key_bindings.vi import vi_bindings
+from prompt_toolkit.key_bindings.emacs import emacs_bindings
 from prompt_toolkit.line import Line
 from prompt_toolkit.prompt import Prompt, TokenList, BracketsMismatchProcessor, PopupCompletionMenu, HorizontalCompletionMenu
+from prompt_toolkit.keys import Key
 
 import jedi
+import platform
 import re
 import sys
 
@@ -81,7 +84,7 @@ class PythonStyle(Style):
         Token.Prompt.ISearch.Text.NoMatch: 'bg:#aa4444 #ffffff',
 
         Token.Prompt.SecondLinePrefix: 'bold #888888',
-        Token.Prompt.LineNumber:       '#bbbbbb',# #ffffff',
+        Token.Prompt.LineNumber:       '#aa6666',
         Token.Prompt.Arg:              'noinherit',
         Token.Prompt.Arg.Text:          'bold',
 
@@ -99,54 +102,69 @@ class PythonStyle(Style):
         Token.CompletionMenu.JediDescription:        'bg:#888888 #cccccc',
         Token.CompletionMenu.CurrentJediDescription: 'bg:#bbbbbb #000000',
 
-        Token.WildMenu.Completion:              'bg:#dddddd #000000',
-        Token.WildMenu.CurrentCompletion:       'bg:#888888 #ffffbb',
-        Token.WildMenu:                         'bg:#dddddd',
-        Token.WildMenu.Arrow:                   'bg:#dddddd #888888',
+        Token.HorizontalMenu.Completion:              '#888888 noinherit',
+        Token.HorizontalMenu.CurrentCompletion:       'bold',
+        Token.HorizontalMenu:                         'noinherit',
+        Token.HorizontalMenu.Arrow:                   'bold #888888',
 
         # Grayed
         Token.Aborted:                 '#aaaaaa',
 
         Token.ValidationError:         'bg:#aa0000 #ffffff',
+
+        # Vi tildes
+        Token.Leftmargin.Tilde:   '#888888',
     }
 
 
-class _PythonInputStreamHandlerMixin(object): # XXX: Don't use Mixins for this. There are better solutions.
+def python_bindings(registry, line):
     """
-    Extensions to the input stream handler for custom 'enter' behaviour.
+    Custom key bindings.
     """
-    def F6(self):
+    handle = registry.add_binding
+
+    @handle(Key.F6)
+    def _(event):
         """
         Enable/Disable paste mode.
         """
-        self.line.paste_mode = not self.line.paste_mode
-        if self.line.paste_mode:
-            self.line.is_multiline = True
+        line.paste_mode = not line.paste_mode
+        if line.paste_mode:
+            line.is_multiline = True
 
-    def F7(self):
+    @handle(Key.F7)
+    def _(event):
         """
         Enable/Disable multiline mode.
         """
-        self.line.is_multiline = not self.line.is_multiline
+        line.is_multiline = not line.is_multiline
 
-    def tab(self):
+    @handle(Key.Tab)
+    def _(event):
         # When the 'tab' key is pressed with only whitespace character before the
         # cursor, do autocompletion. Otherwise, insert indentation.
-        current_char = self.line.document.current_line_before_cursor
+        current_char = line.document.current_line_before_cursor
         if not current_char or current_char.isspace():
-            self.line.insert_text('    ')
+            line.insert_text('    ')
         else:
-            self.line.complete_next()
+            line.complete_next()
+            event.input_processor.input_mode = InputMode.COMPLETE
 
-    def backtab(self):
-        if self.line.mode == LineMode.COMPLETE:
-            self.line.complete_previous()
+    @handle(Key.BackTab, in_mode=InputMode.COMPLETE)
+    def _(event):
+        line.complete_previous()
 
-    def enter(self):
-        self._auto_enable_multiline()
-        super(_PythonInputStreamHandlerMixin, self).enter()
+    @handle(Key.ControlJ) # TODO XXX: !!!!
+    @handle(Key.ControlM)
+    def _(event):
+        _auto_enable_multiline()
 
-    def _auto_enable_multiline(self):
+        if line.is_multiline:
+            line.newline()
+        else:
+            line.return_input()
+
+    def _auto_enable_multiline():
         """
         (Temporarily) enable multiline when pressing enter.
         When:
@@ -155,26 +173,18 @@ class _PythonInputStreamHandlerMixin(object): # XXX: Don't use Mixins for this.
         """
         def is_empty_or_space(s):
             return s == '' or s.isspace()
-        cursor_at_the_end = self.line.document.is_cursor_at_the_end
+        cursor_at_the_end = line.document.is_cursor_at_the_end
 
         # If we just typed a colon, or still have open brackets, always insert a real newline.
-        if cursor_at_the_end and (self.line._colon_before_cursor() or
-                                  self.line._has_unclosed_brackets() or
-                                  self.line._starting_with_at()):
-            self.line.is_multiline = True
+        if cursor_at_the_end and (line._colon_before_cursor() or
+                                  line._has_unclosed_brackets() or
+                                  line._starting_with_at()):
+            line.is_multiline = True
 
         # If the character before the cursor is a backslash (line continuation
         # char), insert a new line.
-        elif cursor_at_the_end and (self.line.document.text_before_cursor[-1:] == '\\'):
-            self.line.is_multiline = True
-
-
-class PythonViInputStreamHandler(_PythonInputStreamHandlerMixin, ViInputStreamHandler):
-    pass
-
-
-class PythonEmacsInputStreamHandler(_PythonInputStreamHandlerMixin, EmacsInputStreamHandler):
-    pass
+        elif cursor_at_the_end and (line.document.text_before_cursor[-1:] == '\\'):
+            line.is_multiline = True
 
 
 class PythonLine(Line):
@@ -264,7 +274,7 @@ class PythonLine(Line):
         """
         before_cursor = self.document.current_line_before_cursor
 
-        if not self.paste_mode and not self.mode == LineMode.INCREMENTAL_SEARCH and before_cursor.isspace():
+        if not self.paste_mode and before_cursor.isspace():
             count = 1 + (len(before_cursor) - 1) % 4
         else:
             count = 1
@@ -283,7 +293,7 @@ class PythonLine(Line):
         # Count space characters, after the cursor.
         after_cursor_space_count = len(after_cursor) - len(after_cursor.lstrip())
 
-        if (not self.paste_mode and not self.mode == LineMode.INCREMENTAL_SEARCH and
+        if (not self.paste_mode and
                     (not before_cursor or before_cursor.isspace()) and after_cursor_space_count):
             count = min(4, after_cursor_space_count)
         else:
@@ -296,42 +306,37 @@ class PythonLine(Line):
 class PythonPrompt(Prompt):
     input_processors = [ BracketsMismatchProcessor() ]
 
-    def __init__(self, render_context, pythonline):
-        super(PythonPrompt, self).__init__(render_context)
-        self._pythonline = pythonline
-
     @property
     def tokens_before_input(self):
-        return [(Token.Prompt, 'In [%s]: ' % self._pythonline.current_statement_index)]
+        return [(Token.Prompt, 'In [%s]: ' % self.commandline.current_statement_index)]
 
     @property
     def completion_menu(self):
-        style = self._pythonline.autocompletion_style
+        style = self.commandline.autocompletion_style
 
         if style == AutoCompletionStyle.POPUP_MENU:
             return PopupCompletionMenu()
         elif style == AutoCompletionStyle.HORIZONTAL_MENU:
-            return HorizontalCompletionMenu()
+            return None
         elif style == AutoCompletionStyle.EXTENDED_POPUP_MENU:
             return ExtendedPopupCompletionMenu()
 
-    def get_help_tokens(self):
+    def write_second_toolbar(self, screen):
         """
         When inside functions, show signature.
         """
-        result = []
-        result.append((Token, '\n'))
-
-        if self.line.mode == LineMode.INCREMENTAL_SEARCH:
-            result.extend(self.isearch_prompt)
-        elif self.line._arg_prompt_text:
-            result.extend(self.arg_prompt)
+        if self.commandline.input_processor.input_mode == InputMode.INCREMENTAL_SEARCH:
+            screen.write_highlighted(list(self.isearch_prompt))
+        elif self.commandline.input_processor.arg is not None:
+            screen.write_highlighted(list(self.arg_prompt))
         elif self.line.validation_error:
-            result.extend(self._get_error_tokens())
+            screen.write_highlighted(list(self._get_error_tokens()))
+        elif self.commandline.autocompletion_style == AutoCompletionStyle.HORIZONTAL_MENU and \
+                        self.line.complete_state and \
+                        self.commandline.input_processor.input_mode == InputMode.COMPLETE:
+            HorizontalCompletionMenu().write(screen, None, self.line.complete_state)
         else:
-            result.extend(self._get_signature_tokens())
-
-        return result
+            screen.write_highlighted(list(self._get_signature_tokens()))
 
     def _get_signature_tokens(self):
         result = []
@@ -354,12 +359,6 @@ class PythonPrompt(Prompt):
 
             result.pop() # Pop last comma
             append((Signature.Operator, ')'))
-
-        append((Signature, ' \n'))
-                            # Note the space before the newline.
-                            # This is to make sure that this line gets at least
-                            # some content in the screen, so that we don't put
-                            # the toolbar here already.
         return result
 
     def _get_error_tokens(self):
@@ -372,95 +371,124 @@ class PythonPrompt(Prompt):
         else:
             return []
 
-    def create_left_input_margin(self, screen, row):
+    def create_left_input_margin(self, screen, row, is_new_line):
         prompt_width = len(TokenList(self.tokens_before_input))
 
-        text = '%i. ' % (row + 1)
+        if is_new_line:
+            text = '%i. ' % row
+        else:
+            text = ''
+
         text = ' ' * (prompt_width - len(text)) + text
 
         screen.write_highlighted([
             (Token.Prompt.LineNumber, text)
         ])
 
-    def write_after_input(self, screen):
-        if not (self.render_context.accept or self.render_context.abort):
-            screen.write_highlighted(list(self.get_help_tokens()))
+    def _get_bottom_position(self, screen, last_screen_height):
+        # Draw the menu at the bottom position.
+        return max(screen.current_height, last_screen_height - 2) - 1
 
-    def write_to_screen(self, screen, last_screen_height):
-        super(PythonPrompt, self).write_to_screen(screen, last_screen_height)
-        self.write_toolbar(screen, last_screen_height)
+    def _print_tildes(self, screen):
+        """
+        Print tildes in the left margin between the last input line and the
+        toolbars. (This is what Vi-users like.)
+        """
+        if self.commandline.vi_mode:
+            last_char_y, last_char_x = screen._cursor_mappings[len(self.line.document.text)]
 
-    def write_toolbar(self, screen, last_screen_height):
-        if not (self.render_context.accept or self.render_context.abort):
-            # Draw the menu at the bottom position.
-            screen._y = max(screen.current_height, last_screen_height - 1) - 1
+            # Fill space with tildes
+            for y in range(last_char_y + 1, screen.current_height - 2):
+                screen.write_at_pos(y, 1, '~', Token.Leftmargin.Tilde)
 
-            result = TokenList()
-            append = result.append
-            TB = Token.Toolbar
+    def write_to_screen(self, screen, last_screen_height, accept=False, abort=False):
+        self.write_before_input(screen)
+        self.write_input(screen)
 
-            append((TB, ' '))
+        if not (accept or abort):
+            self.write_menus(screen)
 
-            # Mode
-            if self.line.mode == LineMode.INCREMENTAL_SEARCH:
-                append((TB.Mode, '(SEARCH)'))
+            y = self._get_bottom_position(screen, last_screen_height)
+
+            screen._y, screen._x = y + 1, 0
+            self.write_second_toolbar(screen)
+
+            screen._y, screen._x = y + 2, 0
+            self.write_toolbar(screen)
+
+            self._print_tildes(screen)
+
+    def write_toolbar(self, screen):
+        TB = Token.Toolbar
+        mode = self.commandline.input_processor.input_mode
+
+        result = TokenList()
+        append = result.append
+
+        append((TB, ' '))
+
+        # Mode
+        if mode == InputMode.INCREMENTAL_SEARCH:
+            append((TB.Mode, '(SEARCH)'))
+            append((TB, '   '))
+        elif self.commandline.vi_mode:
+            if mode == InputMode.VI_NAVIGATION:
+                append((TB.Mode, '(NAV)'))
+                append((TB, '      '))
+            elif mode == InputMode.VI_INSERT:
+                append((TB.Mode, '(INSERT)'))
+                append((TB, '   '))
+            elif mode == InputMode.VI_REPLACE:
+                append((TB.Mode, '(REPLACE)'))
                 append((TB, '  '))
-            elif self._pythonline.vi_mode:
-                mode = self._pythonline._inputstream_handler._vi_mode
-                if mode == ViMode.NAVIGATION:
-                    append((TB.Mode, '(NAV)'))
-                    append((TB, '     '))
-                elif mode == ViMode.INSERT:
-                    append((TB.Mode, '(INSERT)'))
-                    append((TB, '  '))
-                elif mode == ViMode.REPLACE:
-                    append((TB.Mode, '(REPLACE)'))
-                    append((TB, ' '))
-
-                if self._pythonline._inputstream_handler.is_recording_macro:
-                    append((TB.Mode, 'recording'))
-                    append((TB, ' '))
-
-            else:
-                append((TB.Mode, '(emacs)'))
+            elif mode == InputMode.COMPLETE:
+                append((TB.Mode, '(COMPLETE)'))
                 append((TB, ' '))
 
-            # Position in history.
-            append((TB, '%i/%i ' % (self.line._working_index + 1, len(self.line._working_lines))))
+#            if self.commandline._input_processor.is_recording_macro:
+#                append((TB.Mode, 'recording'))
+#                append((TB, ' '))
 
-            # Shortcuts.
-            if self.line.mode == LineMode.INCREMENTAL_SEARCH:
-                append((TB, '[Ctrl-G] Cancel search [Enter] Go to this position.'))
+        else:
+            append((TB.Mode, '(emacs)'))
+            append((TB, ' '))
+
+        # Position in history.
+        append((TB, '%i/%i ' % (self.line._working_index + 1, len(self.line._working_lines))))
+
+        # Shortcuts.
+        if mode == InputMode.INCREMENTAL_SEARCH:
+            append((TB, '[Ctrl-G] Cancel search [Enter] Go to this position.'))
+        else:
+            if self.line.paste_mode:
+                append((TB.On, '[F6] Paste mode (on)  '))
             else:
-                if self.line.paste_mode:
-                    append((TB.On, '[F6] Paste mode (on)  '))
-                else:
-                    append((TB.Off, '[F6] Paste mode (off) '))
+                append((TB.Off, '[F6] Paste mode (off) '))
 
-                if self.line.is_multiline:
-                    append((TB.On, '[F7] Multiline (on)'))
-                else:
-                    append((TB.Off, '[F7] Multiline (off)'))
-
-                if self.line.is_multiline:
-                    append((TB, ' [Meta+Enter] Execute'))
-
-                # Python version
-                version = sys.version_info
-                append((TB, ' - '))
-                append((TB.PythonVersion, 'Python %i.%i.%i' % (version.major, version.minor, version.micro)))
-
-            # Adjust toolbar width.
-            if len(result) > screen.columns:
-                # Trim toolbar
-                result = result[:screen.columns - 3]
-                result.append((TB, ' > '))
+            if self.line.is_multiline:
+                append((TB.On, '[F7] Multiline (on)'))
             else:
-                # Extend toolbar until the page width.
-                result.append((TB, ' ' * (screen.columns - len(result))))
+                append((TB.Off, '[F7] Multiline (off)'))
 
-            screen.write_highlighted([(Token, '\n')])
-            screen.write_highlighted(result)
+            if self.line.is_multiline:
+                append((TB, ' [Meta+Enter] Execute'))
+
+            # Python version
+            version = sys.version_info
+            append((TB, ' - '))
+            append((TB.PythonVersion, '%s %i.%i.%i' % (platform.python_implementation(),
+                                version.major, version.minor, version.micro)))
+
+        # Adjust toolbar width.
+        if len(result) > screen.size.columns:
+            # Trim toolbar
+            result = result[:screen.size.columns - 3]
+            result.append((TB, ' > '))
+        else:
+            # Extend toolbar until the page width.
+            result.append((TB, ' ' * (screen.size.columns - len(result))))
+
+        screen.write_highlighted(result)
 
 
 class ExtendedPopupCompletionMenu(PopupCompletionMenu):
@@ -534,14 +562,12 @@ class PythonCode(Code):
 
 class PythonCommandLine(CommandLine):
     line_factory = PythonLine
-
-    enable_concurency = True
+    prompt_factory = PythonPrompt
 
     def __init__(self, globals=None, locals=None, vi_mode=False, stdin=None, stdout=None, history_filename=None,
                     style=PythonStyle, autocompletion_style=AutoCompletionStyle.POPUP_MENU):
 
         self.globals = globals or {}
-        self.globals.update({k: getattr(__builtins__, k) for k in dir(__builtins__)})
         self.locals = locals or {}
         self.history_filename = history_filename
         self.style = style
@@ -562,23 +588,18 @@ class PythonCommandLine(CommandLine):
             return History()
 
     @property
-    def inputstream_handler_factory(self):
+    def key_bindings_factories(self):
         if self.vi_mode:
-            return PythonViInputStreamHandler
+            return [vi_bindings, python_bindings]
         else:
-            return PythonEmacsInputStreamHandler
-
-    def prompt_factory(self, render_context):
-        # The `PythonPrompt` class needs a reference back in order to show the
-        # input method.
-        return PythonPrompt(render_context, self)
+            return [emacs_bindings, python_bindings]
 
     def code_factory(self, document):
         # The `PythonCode` needs a reference back to this class in order to do
         # autocompletion on the globals/locals.
         return PythonCode(document, self.globals, self.locals)
 
-    def on_input_timeout(self, code_obj):
+    def on_input_timeout(self):
         """
         When there is no input activity,
         in another thread, get the signature of the current code.
@@ -588,6 +609,8 @@ class PythonCommandLine(CommandLine):
             return
         self.get_signatures_thread_running = True
 
+        code_obj = self.line.create_code_obj()
+
         def run():
             script = code_obj._get_jedi_script()
 
@@ -609,10 +632,10 @@ class PythonCommandLine(CommandLine):
 
             # Set signatures and redraw if the text didn't change in the
             # meantime. Otherwise request new signatures.
-            if self._line.text == code_obj.text:
-                self._line.signatures = signatures
+            if self.line.text == code_obj.text:
+                self.line.signatures = signatures
                 self.request_redraw()
             else:
-                self.on_input_timeout(self._line.create_code_obj())
+                self.on_input_timeout()
 
         self.run_in_executor(run)
