@@ -285,6 +285,21 @@ class Document(object):
         except StopIteration:
             pass
 
+    def find_previous_word_beginning(self, count=1):
+        """
+        Return an index relative to the cursor position pointing to the start
+        of the next word. Return `None` if nothing was found.
+        """
+        iterator = _FIND_WORD_RE.finditer(self.text_before_cursor[::-1])
+
+        try:
+            for i, match in enumerate(iterator):
+                if i + 1 == count:
+                    return - match.end(1)
+        except StopIteration:
+            pass
+
+
     def find_next_matching_line(self, match_func): # TODO: unittest.
         """
         Look downwards for empty lines.
@@ -320,40 +335,51 @@ class Document(object):
         Return the relative cursor position (character index) where we would be if the
         user pressed the arrow-up button.
         """
-        if '\n' in self.text_before_cursor:
+        assert count >= 1
+
+        count = min(self.text_before_cursor.count('\n'), count)
+
+        if count:
+            pos = self.cursor_position_col
+
             lines = self.text_before_cursor.split('\n')
-            current_line = lines[-1] # before the cursor
-            previous_line = lines[-2]
+            skip_lines = '\n'.join(lines[-count-1:])
+            new_line = lines[-count-1]
 
             # When the current line is longer then the previous, move to the
             # last character of the previous line.
-            if len(current_line) > len(previous_line):
-                return - len(current_line) - 1
+            if pos > len(new_line):
+                return - len(skip_lines) + len(new_line)
 
             # Otherwise find the corresponding position in the previous line.
             else:
-                return - len(previous_line) - 1
+                return - len(skip_lines) + pos
         return 0
 
-    def get_cursor_down_position(self, count=1): # TODO: implement `count`
+    def get_cursor_down_position(self, count=1):
         """
         Return the relative cursor position (character index) where we would be if the
         user pressed the arrow-down button.
         """
-        if '\n' in self.text_after_cursor:
-            pos = len(self.text_before_cursor.split('\n')[-1])
+        assert count >= 1
+
+        count = min(self.text_after_cursor.count('\n'), count)
+
+        if count:
+            pos = self.cursor_position_col
+
             lines = self.text_after_cursor.split('\n')
-            current_line = lines[0] # after the cursor
-            next_line = lines[1]
+            skip_lines = '\n'.join(lines[:count])
+            new_line = lines[count]
 
             # When the current line is longer then the previous, move to the
             # last character of the next line.
-            if pos > len(next_line):
-                return len(current_line) + len(next_line) + 1
+            if pos > len(new_line):
+                return len(skip_lines) + len(new_line) + 1
 
             # Otherwise find the corresponding position in the next line.
             else:
-                return len(current_line) + pos + 1
+                return len(skip_lines) + pos + 1
 
         return 0
 
@@ -393,6 +419,18 @@ class Document(object):
         """ Relative position for the end of the document. """
         return len(self.text) - self.cursor_position
 
+    def get_start_of_line_position(self, after_whitespace=False):
+        """ Relative position for the end of this line. """
+        if after_whitespace:
+            current_line = self.current_line
+            return len(current_line) - len(current_line.lstrip()) - self.cursor_position_col
+        else:
+            return - len(self.current_line_before_cursor)
+
+    def get_end_of_line_position(self):
+        """ Relative position for the end of this line. """
+        return len(self.current_line_after_cursor)
+
     def get_column_cursor_position(self, column):
         """
         Return the relative cursor position for this column at the current
